structure Tokens = Tokens

type pos = int
type col = int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

val lineNum = ref 1
val pos = ref 1
val col = ref 1

fun advance_column(str) = col := (!col) + String.size(str)

fun read_intlit(str) = 
   let val unbounded = valOf(IntInf.fromString(str)) 
       val warn = (fn(input, min, max, assumed_val) => (
            print("warning: " ^ input ^ " is outside ["^min^","^max^"].\n");
            print("         assuming " ^ assumed_val ^ " instead\n")))
       val toStr = (fn someInt32 => Int32.toString(valOf(someInt32)))
   in
    valOf(
     if unbounded < ~2147483648 then (
       warn(str, 
            toStr(Int32.minInt), 
            toStr(Int32.maxInt), 
            toStr(Int32.minInt));
       Int32.minInt
     )
     else if unbounded > 2147483647 then (
       warn(str, 
            toStr(Int32.minInt), 
            toStr(Int32.maxInt), 
            toStr(Int32.maxInt));
       Int32.maxInt
     ) else (
       Int32.fromString(str)
     ))
   end

fun eof () = Tokens.EOF(!pos, !col)
fun error (e,l : int,_) = TextIO.output (TextIO.stdOut, String.concat[
	"line ", (Int.toString l), ": ", e, "\n"
      ])

val want_debug =  false;

fun debug str = 
    if want_debug then print ("lexer found: "^str^"\n") else ()
(*

STRLIT, CHARLIT regexs taken from ansi c grammar: 
http://www.lysator.liu.se/c/ANSI-C-grammar-l.html
Um. You should be able to explain how those work, mister.
*)

%%
%header (functor CSXLexFun(structure Tokens: CSX_TOKENS));
NEGATIVE=~;
NON_ZERO_DIGIT=[1-9];
DIGIT=[0-9];
ALPHABETIC=[A-Za-z];
ALPHA_NUMERIC=[A-Za-z0-9];
IDENTIFIER={ALPHABETIC}+{ALPHA_NUMERIC}*;

INTLIT=(0?|({NEGATIVE}?{NON_ZERO_DIGIT}{DIGIT}*));
STRLIT=\"(\\.|[^\\"])*\";
CHARLIT='(\\.|[^\\'])';

LINE_COMMENT=\/\/.*;

BLOCK_COMMENT=##(#?[^#])*##;


alpha=[a-zA-Z_][_|a-z|0-9|A-Z]*;
digit=[0-9];
ws=[\ \t];
%%

\n             => (pos := (!pos) + 1; col := 0; debug (yytext^"=>return ");lex());
{ws}+          => (advance_column(yytext); debug (yytext^"=>return ");lex());

[wW][hH][iI][lL][eE]          => (debug (yytext^"=>RW_WHILE "); advance_column(yytext); Tokens.RW_WHILE(!pos, !col));
[bB][oO][oO][lL]           => (debug (yytext^"=>RW_BOOL "); advance_column(yytext); Tokens.RW_BOOL(!pos, !col));
[iI][fF]             => (debug (yytext^"=>RW_IF "); advance_column(yytext); Tokens.RW_IF(!pos, !col));
[eE][xX][iI][tT]           => (debug (yytext^"=>RW_EXIT "); advance_column(yytext); Tokens.RW_EXIT(!pos, !col));
[cC][oO][nN][tT][iI][nN][uU][eE]       => (debug (yytext^"=>RW_CONTINUE "); advance_column(yytext); Tokens.RW_CONTINUE(!pos, !col));
[fF][aA][lL][sS][eE]          => (debug (yytext^"=>RW_FALSE "); advance_column(yytext); Tokens.RW_FALSE(!pos, !col));
[pP][rR][iI][nN][tT]          => (debug (yytext^"=>RW_PRINT "); advance_column(yytext); Tokens.RW_PRINT(!pos, !col));
[tT][rR][uU][eE]           => (debug (yytext^"=>RW_TRUE "); advance_column(yytext); Tokens.RW_TRUE(!pos, !col));
[vV][oO][iI][dD]           => (debug (yytext^"=>RW_VOID "); advance_column(yytext); Tokens.RW_VOID(!pos, !col));
[cC][oO][nN][sS][tT]          => (debug (yytext^"=>RW_CONST "); advance_column(yytext); Tokens.RW_CONST(!pos, !col));
[eE][lL][sS][eE]           => (debug (yytext^"=>RW_ELSE "); advance_column(yytext); Tokens.RW_ELSE(!pos, !col));
[rR][eE][aA][dD]           => (debug (yytext^"=>RW_READ "); advance_column(yytext); Tokens.RW_READ(!pos, !col));
[iI][nN][tT]            => (debug (yytext^"=>RW_INT "); advance_column(yytext); Tokens.RW_INT(!pos, !col));
[tT][hH][eE][nN]           => (debug (yytext^"=>RW_THEN "); advance_column(yytext); Tokens.RW_THEN(!pos, !col));
[cC][lL][aA][sS][sS]          => (debug (yytext^"=>RW_CLASS "); advance_column(yytext); Tokens.RW_CLASS(!pos, !col));
[rR][eE][tT][uU][rR][nN]         => (debug (yytext^"=>RW_RETURN "); advance_column(yytext); Tokens.RW_RETURN(!pos, !col));
[cC][hH][aA][rR]           => (debug (yytext^"=>RW_CHAR "); advance_column(yytext); Tokens.RW_CHAR(!pos, !col));
[bB][rR][eE][aA][kK]          => (debug (yytext^"=>RW_BREAK "); advance_column(yytext); Tokens.RW_BREAK(!pos, !col));

{STRLIT}       => (debug (yytext^"=>STRLIT "); advance_column(yytext); Tokens.STRLIT(yytext, !pos, !col));
{CHARLIT}      => (debug (yytext^"=>CHARLIT "); advance_column(yytext); Tokens.CHARLIT(yytext, !pos, !col));
{INTLIT}+       => (debug (yytext^"=>INTLIT"); 
                    advance_column(yytext);
                    Tokens.INTLIT(read_intlit(yytext), !pos, !col));
{alpha}        => (debug (yytext^"=>ID "); advance_column(yytext); Tokens.ID(yytext, !pos, !col));
"+"            => (debug (yytext^"=>CPLUS "); advance_column(yytext); Tokens.CPLUS(!pos, !col));
"*"            => (debug (yytext^"=>CTIMES "); advance_column(yytext); Tokens.CTIMES(!pos, !col));
";"            => (debug (yytext^"=>SEMI "); advance_column(yytext);Tokens.SEMI(!pos, !col));
"="            => (debug (yytext^"=>ASG "); advance_column(yytext); Tokens.ASG(!pos, !col));
"=="           => (debug (yytext^"=>CEQ "); advance_column(yytext); Tokens.CEQ(!pos, !col));
"!="           => (debug (yytext^"=>CNOTEQ "); advance_column(yytext); Tokens.CNOTEQ(!pos, !col));
">="           => (debug (yytext^"=>CGEQ "); advance_column(yytext); Tokens.CGEQ(!pos, !col));
">"            => (debug (yytext^"=>CGT "); advance_column(yytext); Tokens.CGT(!pos, !col));
"<="           => (debug (yytext^"=>CLEQ "); advance_column(yytext); Tokens.CLEQ(!pos, !col));
"<"            => (debug (yytext^"=>CLT "); advance_column(yytext); Tokens.CLT(!pos, !col));
"-"            => (debug (yytext^"=>CMINUS"); advance_column(yytext); Tokens.CMINUS(!pos, !col));
"^"            => (debug (yytext^"=>CARAT "); advance_column(yytext); Tokens.CARAT(!pos, !col));
"/"            => (debug (yytext^"=>CDIV "); advance_column(yytext); Tokens.CDIV(!pos, !col));
"{"            => (debug (yytext^"=>LBRACE "); advance_column(yytext); Tokens.LBRACE(!pos, !col));
"}"            => (debug (yytext^"=>RBRACE "); advance_column(yytext); Tokens.RBRACE(!pos, !col));
"["            => (debug (yytext^"=>LBRACKET "); advance_column(yytext); Tokens.LBRACKET(!pos, !col));
"]"            => (debug (yytext^"=>RBRACKET "); advance_column(yytext); Tokens.RBRACKET(!pos, !col));
"("            => (debug (yytext^"=>LPAREN "); advance_column(yytext); Tokens.LPAREN(!pos, !col));
")"            => (debug (yytext^"=>RPAREN "); advance_column(yytext); Tokens.RPAREN(!pos, !col));
"\\"           => (debug (yytext^"=>SLASH "); advance_column(yytext); Tokens.SLASH(!pos, !col));
"!"            => (debug (yytext^"=>CNOT "); advance_column(yytext); Tokens.CNOT(!pos, !col));
"&&"           => (debug (yytext^"=>CAND "); advance_column(yytext); Tokens.CAND(!pos, !col));
","            => (debug (yytext^"=>COMMA "); advance_column(yytext); Tokens.COMMA(!pos, !col));
":"            => (debug (yytext^"=>COLON"); advance_column(yytext); Tokens.COLON(!pos, !col));
"||"           => (debug (yytext^"=>COR "); advance_column(yytext); Tokens.COR(!pos, !col));
{BLOCK_COMMENT} => (debug (yytext^"=>block comment"); 
                    let val num_nl = length(List.filter (fn c => c = #"\n") (explode yytext))
                    in
                        pos := (!pos) + num_nl;
                        lex()
                    end);
{LINE_COMMENT} => (debug (yytext^"=>line comment");
                   lex());

.      => (error ("Invalid Token: "^yytext, !pos,!pos); 
           OS.Process.exit(OS.Process.success));
