open SymbolTable;

(* So here's the deal on testing the symbol table.
 * Do one thing, and perform a number of assertions as necessary. 
 *)

exception TestFailure

fun assert(expr, message) = 
      if expr then
        ()
      else
        (print(message);
         raise TestFailure)

fun test_insert(table) = 
  (insert_symb("foo", NoneSymbInfo, table);
   assert(get_symb("foo", table) = NoneSymbInfo, "a symb could not be retrieved"));

fun test_get(table, name) = 
   assert(get_symb(name, table) = NoneSymbInfo, "a symb could not be retrieved");

fun negative_test_get_symb(table) = 
   assert(let val do_not_want = get_symb("not there", table) 
          in 
            false
          end
          handle SymbolTable.NotFound => true, 
       "get_symb didn't raise NotFound when it was supposed to.");

fun test_open_scope(table) =
    (let val aug_table = open_scope(table) in 
       test_get(aug_table, "foo")
     end)

fun test_close_scope(table) =
    (close_scope(table);
     test_get(table, "foo"));

fun test_close_too_many_scopes(table) = 
   assert(let val do_not_want = close_scope(close_scope(table))
          in 
            false
          end
          handle SymbolTable.NoAvailableScope => true, 
       "there should have been NoAvailableScope.");

fun test_close_and_get(table) = 
   assert(let val do_not_want = test_get(close_scope(table), "foo")
          in 
            false
          end
          handle SymbolTable.NotFound => true, 
       "the symbol should not have been found here.");

fun main() = 
   let val table = make_symbol_table()
   in
      (test_insert(table);
       negative_test_get_symb(table);
       test_open_scope(table);
       negative_test_get_symb(table);
       test_close_scope(table);
       test_close_and_get(table);
      "All tests passed") 
   end;

main();
